list(
  # Adjustment name ----
  Name = "Reporting Delays",

  # Adjustment type ----
  Type = "REPORTING_DELAYS",

  # Adjustment subtype ----
  SubType = "DEFAULT",

  # Input parameters to the adjustment function ----
  Parameters = list(
    startYear = list(
      label = "Diagnosis start year",
      value = 2000L,
      input = "numeric"),
    endYear = list(
      label = "Notification end year",
      value = 2017,
      input = "numeric"),
    endQrt = list(
      label = "Notification end quarter (integer between 1 and 4)",
      value = 1,
      min = 1,
      max = 4,
      input = "numeric"),
    stratGender = list(
      name = "stratGender",
      label = "Gender",
      value = FALSE,
      input = "checkbox"),
    stratTrans = list(
      name = "stratTrans",
      label = "Transmission",
      value = FALSE,
      input = "checkbox"),
    stratMigr = list(
      name = "stratMigr",
      label = "Migration",
      value = FALSE,
      input = "checkbox")
  ),

  # Names of packages that must be made available to the adjustment function ----
  RequiredPackageNames = c(),

  ## Adjustment function ----
  AdjustmentFunction = function(inputData, parameters) {

    require(data.table)
    require(survival)

    # A) SETUP -------------------------------------------------------------------------------------

    # Work on a copy
    compData <- copy(inputData)

    # Start year
    startYear <- parameters$startYear
    # End quarter
    endQrt <- parameters$endYear + parameters$endQrt / 4
    # Stratifiation columns
    stratVarNames <- c()
    if (parameters$stratGender) {
      stratVarNames <- union(stratVarNames, "Gender")
    }
    if (parameters$stratTrans) {
      stratVarNames <- union(stratVarNames, "Transmission")
    }
    if (parameters$stratMigr) {
      stratVarNames <- union(stratVarNames, "GroupedRegionOfOrigin")
    }
    stratVarNames <- stratVarNames[stratVarNames %in% colnames(compData)]

    # B) PROCESS DATA ------------------------------------------------------------------------------

    # Add dummy "Imputation" column if not found
    isOriginalData <- !("Imputation" %in% colnames(compData))
    if (isOriginalData) {
      compData[, Imputation := 0L]
    }

    # Make sure the strata columns exist in the data
    stratVarNames <- stratVarNames[stratVarNames %in% colnames(compData)]
    stratVarNamesTrend <- union("DateOfDiagnosisYear",
                                stratVarNames)

    # Create intermediate variables
    compData[, ":="(
      NotificationTime = DateOfNotificationYear + 1/4 * DateOfNotificationQuarter,
      DiagnosisTime = DateOfDiagnosisYear + 1/4 * DateOfDiagnosisQuarter
    )]

    # Create dimensions to match the weights later
    outputData <- copy(compData)
    outputData[, MaxNotificationTime := max(NotificationTime, na.rm = TRUE),
               by = .(ReportingCountry)]
    outputData[, VarT := 4 * (pmin(MaxNotificationTime, endQrt) - DiagnosisTime) + 1]

    # Filter
    compData <- compData[!is.na(DiagnosisTime) & !is.na(NotificationTime)]
    compData <- compData[DiagnosisTime >= max(startYear + 0.25,
                                              min(NotificationTime, na.rm = TRUE)) &
                           NotificationTime <= endQrt]
    compData[, VarX := 4 * (NotificationTime - DiagnosisTime)]
    compData[VarX < 0, VarX := NA]
    compData[, ":="(
      MinNotificationTime = min(NotificationTime, na.rm = TRUE),
      MaxNotificationTime = max(NotificationTime, na.rm = TRUE)
    ), by = .(ReportingCountry)]
    compData[, ":="(
      VarT = 4 * (pmin(MaxNotificationTime, endQrt) - DiagnosisTime) + 1,
      Tf = 4 * (pmin(MaxNotificationTime, endQrt) - pmax(MinNotificationTime, startYear)),
      ReportingDelay = 1L
    )]
    compData[, ":="(
      VarXs = Tf - VarX,
      VarTs = Tf - VarT
    )]
    # NOTE: Otherwise survival model complains
    compData <- compData[VarXs > VarTs]

    totalPlot <- NULL
    totalPlotData <- NULL
    stratPlotList <- NULL
    stratPlotListData <- NULL
    rdDistribution <- NULL
    reportTableData <- NULL
    univAnalysis <- NULL
    if (nrow(compData) > 0) {
      # ------------------------------------------------------------------------
      # Prepare diagnostic table based on original data
      model <- compData[Imputation == 0L,
                        Surv(time = VarTs,
                             time2 = VarXs,
                             event = ReportingDelay)]

      # Defining univariate models
      univFormulas <- lapply(
        stratVarNamesTrend,
        function(x) as.formula(sprintf("model ~ %s", x)))

      # Applying univariate models
      univModels <- lapply(
        univFormulas,
        function(x) coxph(x, data = compData[Imputation == 0L]))

      # Extract results of univariable analysis (whether particular covariates
      # are associated with RD)
      univAnalysis <- rbindlist(lapply(
        univModels,
        function(x) {
          y <- summary(x)
          z <- cox.zph(x)
          res <- merge(as.data.table(y$conf.int),
                       as.data.table(y$coefficients))
          res <- cbind(res,
                       as.data.table(z$table[rownames(z$table) != "GLOBAL", "p", drop = FALSE]))
          res[, lapply(.SD, signif, 2), .SDcols = colnames(res)]
          setnames(res, c("HR", "1/HR", "HR.lower.95",
                          "HR.upper.95", "Beta", "SE.Beta",
                          "Z", "P.value", "Prop.assumpt.p"))
          res <- cbind(Predictor = rownames(y$conf.int),
                       res)
          return(res)}))

      # ------------------------------------------------------------------------
      # RD estimation with time trend

      # Define parameters
      lastYear <- compData[, max(DateOfDiagnosisYear)]
      years <- (lastYear - 4):lastYear
      tGroups <- 1:3
      imputations <- compData[, sort(unique(Imputation))]
      formula <- as.formula(sprintf("Surv(time = VarTs, time2 = VarXs, event = ReportingDelay) ~ (%s):strata(tgroup)",
                                    paste(stratVarNamesTrend, collapse = " + ")))
      cuts <- compData[, c(max(VarXs) - 10,
                           max(VarXs) - 3)]

      # Run fitting per imputation separately
      fitStratum <- list()
      for (imputation in imputations) {
        compDataSplit <- setDT(survSplit(
          formula = Surv(time = VarTs, time2 = VarXs, event = ReportingDelay) ~ .,
          data = compData[Imputation == imputation],
          cut = cuts,
          episode = "tgroup"))

        fitCox <- coxph(formula,
                        data = compDataSplit)

        estFrame <- CJ(DateOfDiagnosisYear = years,
                       tgroup = tGroups)

        estCov <- na.omit(unique(compData[, ..stratVarNames]))
        if (nrow(estCov) > 0) {
          estFrame[, MergeDummy := 1]
          estCov[, MergeDummy := 1]
          estFrame <- estFrame[estCov,
                               on = .(MergeDummy),
                               allow.cartesian = TRUE]
          estFrame[, MergeDummy := NULL]
        }

        estFrame[, ":="(
          Id = rep(seq_len(.N/length(tGroups)), each = length(tGroups)),
          VarTs = ifelse(tgroup == 1,  0, ifelse(tgroup == 2, 52, ifelse(tgroup == 3, 59, NA))),
          VarXs = ifelse(tgroup == 1, 52, ifelse(tgroup == 2, 59, ifelse(tgroup == 3, 62, NA))),
          ReportingDelay = 0
        )]

        fit <- survfit(fitCox,
                       newdata = estFrame,
                       id = Id)

        fitStratumImp <- data.table(
          Imputation = imputation,
          Delay = fit$time,
          P = fit$surv,
          Weight = 1/fit$surv,
          Var = fit$std.err^2,
          unique(estFrame[, ..stratVarNamesTrend])[rep(seq_len(.N), fit$strata)])
        fitStratumImp[, VarT := max(Delay) - Delay]

        # Store this imputation results
        fitStratum[[as.character(imputation)]] <- fitStratumImp[VarT >= 0]
      }
      fitStratum <- rbindlist(fitStratum)

      # Merge P, Weight and Var with outputData object
      mergeVars <- union(stratVarNamesTrend,
                         c("VarT", "Imputation"))
      outputData[, ":="(
        DateOfDiagnosisYearOrig = DateOfDiagnosisYear,
        DateOfDiagnosisYear = pmax.int(lastYear - 4, DateOfDiagnosisYear),
        Source = ifelse(Imputation == 0, "Reported", "Imputed")
      )]
      outputData <- merge(outputData,
                          fitStratum[, c(..mergeVars, "P", "Weight", "Var")],
                          by = mergeVars,
                          all.x = TRUE)
      outputData[, MissingData := is.na(Weight)]
      outputData[is.na(Weight), Weight := 1]
      outputData[is.na(P), P := 1]
      outputData[is.na(Var), Var := 0]
      outputData[, ":="(
        DateOfDiagnosisYear = DateOfDiagnosisYearOrig,
        DateOfDiagnosisYearOrig = NULL
      )]

      # ------------------------------------------------------------------------

      # Get distribution object as artifact
      varNames <- setdiff(colnames(fitStratum),
                          c("Delay", "P", "Weight", "Var", "VarT"))
      rdDistribution <- fitStratum[VarT > 0,
                                   union(varNames, c("VarT", "P", "Weight", "Var")),
                                   with = FALSE]
      setnames(rdDistribution,
               old = "VarT",
               new = "Quarter")
      setorderv(rdDistribution, union(varNames, "Quarter"))

      # Aggregate and keep only required dimensions
      agregat <- outputData[, .(Count = .N,
                                P = mean(P),
                                Weight = mean(Weight),
                                Var = mean(Var)),
                            by = c(mergeVars, "Source", "MissingData")]

      # Compute estimated count and its variance
      agregat[, ":="(
        EstCount = Count * Weight,
        EstCountVar = (Count * (Count + 1) / P^4 * Var) + Count * (1 - P) / P^2
      )]

      # C) TOTAL PLOT ----------------------------------------------------------
      totalPlotData <- GetRDPlotData(data = agregat,
                                     by = c("MissingData", "Source", "Imputation",
                                            "DateOfDiagnosisYear"))
      setorderv(totalPlotData, c("MissingData", "DateOfDiagnosisYear"))
      totalPlot <- GetRDPlots(plotData = totalPlotData[MissingData == FALSE],
                              isOriginalData = isOriginalData)

      reportTableData <-
        totalPlotData[Source == "Reported" & MissingData == FALSE,
                      .(DateOfDiagnosisYear, Reported = Count, EstCount,
                        LowerEstCount, UpperEstCount)]
      missRdData <-
        totalPlotData[Source == "Reported" & MissingData == TRUE,
                      .(DateOfDiagnosisYear, Count)]
      reportTableData[missRdData,
                      MissingData := Count,
                      on = .(DateOfDiagnosisYear)]
      reportTableData[, ":="(
        EstUnreported = na.zero(EstCount) - na.zero(Reported),
        LowerEstUnreported = na.zero(LowerEstCount) - na.zero(Reported),
        UpperEstUnreported = na.zero(UpperEstCount) - na.zero(Reported)
      )]
      setcolorder(reportTableData,
                  c("DateOfDiagnosisYear", "MissingData", "Reported",
                    "EstUnreported", "EstCount", "LowerEstCount", "UpperEstCount"))

      # D) STRATIFIED PLOT (OPTIONAL) ------------------------------------------
      if (length(stratVarNames) > 0) {
        # Stratification
        colNames <- union(c("MissingData", "Source", "DateOfDiagnosisYear", "Count", "EstCount",
                            "EstCountVar"),
                          c(stratVarNamesTrend, "Imputation"))
        # Keep only required columns, convert data to "long" format...
        agregatLong <- melt(agregat[, ..colNames],
                            measure.vars = stratVarNames,
                            variable.name = "Stratum",
                            value.name = "StratumValue")

        stratPlotListData <- GetRDPlotData(data = agregatLong,
                                           by = c("MissingData", "Source", "Imputation",
                                                  "DateOfDiagnosisYear", "Stratum", "StratumValue"))
        stratPlotList <- lapply(stratVarNames,
                                GetRDPlots,
                                plotData = stratPlotListData[MissingData == FALSE],
                                isOriginalData = isOriginalData)

        names(stratPlotList) <- stratVarNames
      }
    } else {
      outputData[, Weight := 1]
    }

    # Keep only columns present in the input object plus the weight
    outColNames <- union(colnames(inputData),
                         c("VarT", "Weight"))
    outputData <- outputData[, ..outColNames]

    artifacts <- list(OutputPlotTotal = totalPlot,
                      OutputPlotTotalData = totalPlotData,
                      OutputPlotStrat = stratPlotList,
                      OutputPlotStratData = stratPlotListData,
                      ReportTableData = reportTableData,
                      RdDistribution = rdDistribution,
                      UnivAnalysis = univAnalysis)

    cat("No adjustment specific text outputs.\n")

    return(list(Table = outputData,
                Artifacts = artifacts))
  }
)
